from efootprint.constants.units import u

ECOLOGITS_UNIT_MAPPING ={
      "tps": u.dimensionless,
      "ttft": u.s,
      "batch_size": u.dimensionless,
      "datacenter_pue": u.dimensionless,
      "generation_latency": u.s,
      "gpu_embodied_gwp": u.kg,
      "gpu_energy": u.kWh,
      "gpu_energy_alpha": u.Wh,
      "gpu_energy_beta": u.dimensionless,
      "gpu_energy_gamma": u.Wh,
      "gpu_memory": u.GB,
      "gpu_required_count": u.dimensionless,
      "if_electricity_mix_gwp": u.kg / u.kWh,
      "latency_alpha": u.s,
      "latency_beta": u.s,
      "latency_gamma": u.s,
      "model_active_parameter_count": u.billion,
      "model_quantization_bits": u.dimensionless,
      "model_required_memory": u.GB,
      "model_total_parameter_count": u.billion,
      "output_token_count": u.dimensionless,
      "request_embodied_gwp": u.kg,
      "request_energy": u.kWh,
      "request_latency": u.s,
      "request_usage_gwp": u.kg,
      "server_embodied_gwp": u.kg,
      "server_energy": u.kWh,
      "server_gpu_count": u.dimensionless,
      "server_gpu_embodied_gwp": u.kg,
      "server_lifetime": u.s,
      "server_power": u.kW,
  }